/*
   https://en.wikipedia.org/wiki/Barrier_(computer_science)
   https://en.wikipedia.org/wiki/Synchronous_rendezvous
*/

import java.util.concurrent.locks.ReentrantLock;
import java.util.concurrent.locks.Condition;
import java.util.concurrent.TimeUnit;
import java.util.Random;

/**
   Implement this barrier example of thread synchronization.

     thread1   thread2   thread3
            \     |     /
             \    |    /
              \   |   /
               barrier
              /   |   \
             /    |    \
            /     |     \
     thread1   thread2    thread3

   In this example we implement the barrier synchronization
   ourselves using a condition variable.
*/
public class BarrierExample_v2
{
   private static ReentrantLock lock = new ReentrantLock();
   private static Condition cv = lock.newCondition();
   private static int count = 0;

   public static void main(String[] args)
   {
      final Thread t1 = new Thread(new Task(1, 1));
      final Thread t2 = new Thread(new Task(2, 2));
      final Thread t3 = new Thread(new Task(3, 3));
      t1.start();
      t2.start();
      t3.start();
   }


   static class Task implements Runnable
   {
      private final int id;
      private final int colunmNumber;
      private final Random random;

      Task(int id, int columnNumber)
      {
         this.id = id;
         this.colunmNumber = columnNumber;
         this.random = new Random();
      }

      @Override
      public void run()
      {
         // Tab over to the column used by this task.
         final String indent = "\t".repeat(colunmNumber); // Java 11.

         // How long this task should "work".
         int computeTime = 2 + random.nextInt(8); // Between 2 and 10 seconds.

         // Print a message in the appropriate column for this task.
         // The message shows this task's id and the number of seconds
         // left for this task to execute.
         System.out.println(indent + "t" + id + "," + computeTime);
         while (computeTime > 0)
         {
            try{ TimeUnit.SECONDS.sleep(1); }catch(InterruptedException e){}
            --computeTime;
            System.out.println(indent + "t" + id + "," + computeTime);
         }

         // Wait for the other two threads to catch up.
         // Implement our own barrier using a condition variable.
         try
         {
            lock.lock();
            ++count;
            if (count < 3)
            {
               cv.await();
            }
            else
            {
               System.out.println("============================");
               cv.signalAll();
            }
         }
         catch (InterruptedException e)
         {
            System.out.println(e);
         }
         finally
         {
            lock.unlock();
         }

         // How long this task should "work" again.
         computeTime = 2 + random.nextInt(8); // Between 2 and 10 seconds.

         System.out.println(indent + "t" + id + "," + computeTime);
         while (computeTime > 0)
         {
            try{ TimeUnit.SECONDS.sleep(1); }catch(InterruptedException e){}
            --computeTime;
            System.out.println(indent + "t" + id + "," + computeTime);
         }
      }
   }
}
